import numpy as n
import os
from time import time, asctime, localtime
from numpy.random import randn, rand
from numpy import s_, dot, tile, zeros, ones, zeros_like, array, ones_like
from util import *
from data import *
from options import *
from gpumodel import *
from math import ceil, floor, sqrt
import sys
import shutil
import platform

class ConvRBM(GPUModel):
    MODEL_NAME = "convrbm"
    def __init__(self, op, load_dic):
        filename_options = [('filter_size', 'f'), ('num_filters', 'h'), ('num_hid_glob', 'hg'), ('local_max_size', 'max'),
                            ('tied_biases', 'tb'), ('fixed_border_size', 'fb'), ('zero_padding', 'zp'), ('cd', 'cd')]
        GPUModel.__init__(self, self.MODEL_NAME, op, load_dic, filename_options)

    def init_model_lib(self):
        self.libmodel.initModel(self.hid_vis, self.vis_biases, self.hid_biases,
                                self.hid_vis_inc, self.vis_bias_inc, self.hid_bias_inc,
                                self.hid_means,
                                self.vis_hid_glob, self.hid_glob_biases, self.vis_hid_glob_inc, self.hid_glob_bias_inc, 
                                self.hid_glob_means,
                                self.neg_data_pcd,
                                self.fixed_border_size, self.local_max_size, self.zero_padding,
                                self.eps_w, self.eps_vb, self.eps_hb,
                                self.eps_w_glob, self.eps_hb_glob,
                                self.mom, self.weight_cost, 
                                self.sparse_cost, self.sparse_target, 
                                self.sparse_cost_glob, self.sparse_target_glob, 
                                self.minibatch_size, self.cd,
                                self.device_id)
    
    def init_model_state(self):
        self.num_vis = self.train_data_provider.get_data_dims()
        self.filter_pixels = self.filter_size ** 2
        self.img_size = int(sqrt(self.num_vis / 3))
        self.img_pixels = self.img_size ** 2
        self.img_size_padded = self.img_size + 2 * self.zero_padding
        self.conv_outputs_x = self.img_size_padded - self.filter_size + 1
        self.conv_outputs =  self.conv_outputs_x ** 2
        
        assert self.num_vis % 3 == 0
        assert self.img_size - 2 * self.fixed_border_size > 0
        assert self.local_max_size == 0 or self.conv_outputs_x % self.local_max_size == 0
        assert self.cd > 0 or self.fixed_border_size == 0, "Can't keep fixed border with PCD"
        assert self.fixed_border_size == 0 or self.num_hid_glob == 0, "Global units only implemented when there's no fixed border"
        
        if not self.load_dic:
            ms = self.model_state
            ms["hid_vis"] = self.init_weight_mult * randn(self.num_filters, self.filter_pixels * 3).astype(n.single)
            if self.tied_biases:
                ms["hid_biases"] = 0*ones((self.num_filters, 1), dtype=n.single)
            else:
                ms["hid_biases"] = 0*ones((self.num_filters, self.conv_outputs), dtype=n.single)
            ms["vis_biases"] = zeros((1, self.num_vis), dtype=n.single)
            ms["hid_vis_inc"] = zeros_like(ms["hid_vis"])
            ms["hid_bias_inc"] = zeros_like(ms["hid_biases"])
            ms["vis_bias_inc"] = zeros_like(ms["vis_biases"])
            ms["hid_means"] = self.sparse_target * ones((self.num_filters, self.conv_outputs), dtype=n.single)
            
            ms["vis_hid_glob"] = self.init_weight_mult * array(randn(self.num_vis, self.num_hid_glob), dtype=n.single, order='F')
            ms["hid_glob_biases"] = zeros((1, self.num_hid_glob), dtype=n.single, order='F')
            ms["vis_hid_glob_inc"] = array(zeros_like(ms["vis_hid_glob"]), order='F')
            ms["hid_glob_bias_inc"] = array(zeros_like(ms["hid_glob_biases"]), order='F')
            ms["hid_glob_means"] = self.sparse_target_glob * ones_like(ms["hid_glob_biases"])
            
            ms["neg_data_pcd"] = array(randn(self.minibatch_size, self.num_vis), dtype=n.single)
    
    def parse_batch_data(self, batch_data, train=True):
        return batch_data[0], batch_data[1], batch_data[2][0] / self.sigma 
        
    def print_test_results(self):
        GPUModel.print_test_results(self)
        print "\n\t\tWeights: %.6f" % n.abs(self.hid_vis).mean()
        print "\t\tGlobal weights: %.6f" % n.abs(self.vis_hid_glob).mean()
        print "\t\tHid means: %.6f" % self.hid_means.mean()
        print "\t\tGlobal hid means: %.6f" % self.hid_glob_means.mean(),

if __name__ == "__main__":
    op = GPUModel.get_default_options_parser()
    op.add_option("num-filters", "num_filters", IntegerOptionParser, "Number of filters", set_once=True)
    op.add_option("num-globs", "num_hid_glob", IntegerOptionParser, "Number of global hiddens", set_once=True, default=0)
    op.add_option("filter-size", "filter_size", IntegerOptionParser, "Filter size", set_once=True)
    op.add_option("fixed-border-size", "fixed_border_size", IntegerOptionParser, "Fixed border size", set_once=True, default=0)
    op.add_option("local-max-size", "local_max_size", IntegerOptionParser, "Local max size", set_once=True, default=0)
    op.add_option("zero-padding", "zero_padding", IntegerOptionParser, "Zero padding size", set_once=True, default=0)
    op.add_option("weight-mult", "init_weight_mult", FloatOptionParser, "Initial weight stdev", set_once=True, default=0.001)
    op.add_option("tied-biases", "tied_biases", BooleanOptionParser, "Share hidden biases", set_once=True, default=0)
    
    op.add_option("eps-w", "eps_w", FloatOptionParser, "Learning rate: weights", default=0.003)
    op.add_option("eps-hid", "eps_hb", FloatOptionParser, "Learning rate: hidden biases", default=OptionExpression("eps_w * 10"))
    op.add_option("eps-vis", "eps_vb", FloatOptionParser, "Learning rate: visible biases", default=OptionExpression("eps_w * 10"))
    op.add_option("eps-w-glob", "eps_w_glob", FloatOptionParser, "Learning rate: global weights", default=0.00001)
    op.add_option("eps-hid-glob", "eps_hb_glob", FloatOptionParser, "Learning rate: global hidden biases", default=0.01)
    op.add_option("mom", "mom", FloatOptionParser, "Momentum (biases only)", default=0.9)
    op.add_option("weight-cost", "weight_cost", FloatOptionParser, "Weight decay", default=0)
    op.add_option("sigma", "sigma", FloatOptionParser, "Standard deviation for visible units")
    op.add_option("mini", "minibatch_size", IntegerOptionParser, "Minibatch size", default=128)
    op.add_option("cd", "cd", IntegerOptionParser, "CD iterations (0=PCD)", default=1)
    op.add_option("sparse-cost", "sparse_cost", FloatOptionParser, "Sparseness: cost", default=0)
    op.add_option("sparse-target", "sparse_target", FloatOptionParser, "Sparseness: target", default=0.05)
    op.add_option("sparse-cost-glob", "sparse_cost_glob", FloatOptionParser, "Sparseness for globals: cost", default=0)
    op.add_option("sparse-target-glob", "sparse_target_glob", FloatOptionParser, "Sparseness for globals: target", default=0.05)
    
    op.options["max_test_err"].default = 1.5
    op.options["data_path"].default = "/ais/gobi1/kriz/tiny/tinyimages-batches_del1000-batchsize-8064/"
    op.options["test_batch_range"].default = [199]
    op.options["train_batch_range"].default = range(1,199)
    op.options["save_path"].default = "/ais/gobi2/kriz/tiny/convrbm_saves/"
    op.options["sigma"].default = 69
    
    op, load_dic =  GPUModel.parse_options(op)
    model = ConvRBM(op, load_dic)
    model.train()
